from vit.formatter import Formatter
from vit.util import unicode_len

class Tags(Formatter):
    def format(self, tags, task):
        if not tags:
            return self.markup_none(self.colorizer.tag_none())
        elif len(tags) == 1:
            tag = list(tags)[0]
            return (unicode_len(tag), self.markup_element(tag))
        else:
            last_tag = list(tags)[-1]
            width = 0
            text_markup = []
            for tag in tags:
                width += unicode_len(tag)
                text_markup += [self.markup_element(tag)]
                if tag != last_tag:
                    width += 1
                    text_markup += [',']
            return (width, text_markup)

    def colorize(self, tag):
        return self.colorizer.tag(tag)



import unittest

from vit.color_mappings import task_256_to_urwid_256


class TestColorMappings(unittest.TestCase):

    def test_maps_gray_and_grey_base_name(self):
        mapping = task_256_to_urwid_256()
        self.assertEqual(mapping['gray'], 'light gray')
        self.assertEqual(mapping['grey'], 'light gray')

    def test_maps_grey_scale_variants(self):
        mapping = task_256_to_urwid_256()
        self.assertEqual(mapping['gray10'], mapping['grey10'])
        self.assertEqual(mapping['grey10'], 'g40')


if __name__ == '__main__':
    unittest.main()



from vit.formatter import String
from vit.util import unicode_len

class Project(String):
    def __init__(self, column, report, defaults, blocking_task_uuids, **kwargs):
        super().__init__(column, report, defaults, blocking_task_uuids)
        self.indent_subprojects = self.is_subproject_indentable()

    def format(self, project, task):
        return self.format_project(project, task) if project else self.markup_none(self.colorizer.project_none())

    def format_project(self, project, task):
        return self.format_subproject_indented(project, task) if self.indent_subprojects else (unicode_len(project), self.markup_element(project))

    def format_subproject_indented(self, project, task):
        parts = project.split('.')
        (width, spaces, marker, subproject) = self.formatter.format_subproject_indented(parts)
        return (width, [spaces, marker, (self.colorize(project), subproject)])

    def is_subproject_indentable(self):
        return self.formatter.config.subproject_indentable and self.report['subproject_indentable']

    def colorize(self, project):
        return self.colorizer.project(project)



import re
from functools import cmp_to_key, wraps

from vit.color_mappings import task_256_to_urwid_256, task_bright_to_color

VALID_COLOR_MODIFIERS = [
    "bold",
    "underline",
]

INVALID_COLOR_MODIFIERS = [
    "inverse",
]


class TaskColorConfig:
    """Colorized task output."""

    def __init__(self, config, task_config, theme, theme_alt_backgrounds):
        self.config = config
        self.task_config = task_config
        self.theme = theme
        self.theme_alt_backgrounds = theme_alt_backgrounds
        self.include_subprojects = self.config.get("color", "include_subprojects")
        self.task_256_to_urwid_256 = task_256_to_urwid_256()
        # NOTE: Because Taskwarrior disables color on piped commands, and I don't
        # see any portable way to get output from a system command in Python
        # without pipes, the 'color' config setting in Taskwarrior is not used, and
        # instead a custom setting is used.
        self.color_enabled = self.config.get("color", "enabled")
        self.display_attrs_available, self.display_attrs = self.convert_color_config(
            self.task_config.filter_to_dict(r"^color\.")
        )
        self.project_display_attrs = self.get_project_display_attrs()
        if self.include_subprojects:
            self.add_project_children()
        self.inject_alt_background_display_attrs()

    def inject_alt_background_display_attrs(self):
        for display_attr in self.display_attrs.copy():
            (
                name,
                display_attr_foreground_16,
                display_attr_background_16,
                _mono,
                display_attr_foreground_256,
                display_attr_background_256,
            ) = display_attr
            for modifier, alt_backgrounds in self.theme_alt_backgrounds.items():
                display_attr_modifier = name + modifier
                if self.has_display_attr(name):
                    self.display_attrs_available[display_attr_modifier] = True
                    alt_background_16, alt_background_256 = alt_backgrounds
                    new_background_16 = (
                        alt_background_16
                        if display_attr_background_16 == ""
                        else display_attr_background_16
                    )
                    new_background_256 = (
                        alt_background_256
                        if display_attr_background_256 == ""
                        else display_attr_background_256
                    )
                    self.display_attrs.append(
                        self.make_display_attr(
                            display_attr_modifier,
                            display_attr_foreground_256,
                            new_background_256,
                            foreground_16=display_attr_foreground_16,
                            background_16=new_background_16,
                        )
                    )
                else:
                    self.display_attrs_available[display_attr_modifier] = False

    def add_project_children(self):
        color_prefix = "color.project."
        for display_attr, fg16, bg16, m, fg256, bg256 in self.project_display_attrs:
            for entry in self.task_config.projects:
                attr = "%s%s" % (color_prefix, entry)
                if not self.has_display_attr(attr) and attr.startswith(
                    "%s." % display_attr
                ):
                    self.display_attrs_available[attr] = True
                    self.display_attrs.append((attr, fg16, bg16, m, fg256, bg256))

    def has_display_attr(self, display_attr):
        return (
            display_attr in self.display_attrs_available
            and self.display_attrs_available[display_attr]
        )

    def get_project_display_attrs(self):
        return sorted(
            [
                (a, fg16, bg16, m, fg256, bg256)
                for (a, fg16, bg16, m, fg256, bg256) in self.display_attrs
                if self.display_attrs_available[a] and self.is_project_display_attr(a)
            ],
            reverse=True,
        )

    def is_project_display_attr(self, display_attr):
        return display_attr[0:14] == "color.project."

    def convert_color_config(self, color_config):
        display_attrs_available = {}
        display_attrs = []
        for key, value in color_config.items():
            foreground, background = self.convert_colors(value)
            available = self.has_color_config(foreground, background)
            display_attrs_available[key] = available
            if available:
                display_attrs.append(
                    self.make_display_attr(key, foreground, background)
                )
        return display_attrs_available, display_attrs

    def make_display_attr(
        self,
        display_attr,
        foreground,
        background,
        foreground_16=None,
        background_16=None,
    ):
        # TODO: 256 colors need to be translated down to 16 color mode.
        foreground_16 = "" if foreground_16 is None else foreground_16
        background_16 = "" if background_16 is None else background_16
        return (display_attr, foreground_16, background_16, "", foreground, background)

    def has_color_config(self, foreground, background):
        return foreground != "" or background != ""

    def convert_colors(self, color_config):
        # TODO: Maybe a fancy regex eventually...
        color_config = task_bright_to_color(color_config).strip()
        starts_with_on = color_config[0:3] == "on "
        parts = list(map(lambda p: p.strip(), color_config.split("on ")))
        foreground, background = (
            (parts[0], parts[1])
            if len(parts) > 1
            else (None, parts[0])
            if starts_with_on
            else (parts[0], None)
        )
        foreground_parts, background_parts = self.make_color_parts(
            foreground, background
        )
        return self.convert_color_parts(foreground_parts), self.convert_color_parts(
            background_parts
        )

    def convert_color_parts(self, color_parts):
        sorted_parts = self.sort_color_parts(color_parts)
        remapped_colors = self.map_named_colors(sorted_parts)
        return ",".join(remapped_colors)

    def check_invalid_color_parts(self, color_parts):
        invalid_color_parts = {*color_parts} & {*INVALID_COLOR_MODIFIERS}
        if invalid_color_parts:
            color_parts = [
                color for color in color_parts if color not in invalid_color_parts
            ]
            print(
                "WARNING: The following TaskWarrior color definitions are unsupported in VIT, and have been removed from the definition: %s -- read the documentation for possible workarounds"
                % ", ".join(invalid_color_parts)
            )
        return color_parts

    def map_named_colors(self, color_parts):
        if len(color_parts) > 0 and color_parts[0] in self.task_256_to_urwid_256:
            color_parts[0] = self.task_256_to_urwid_256[color_parts[0]]
        return color_parts

    def make_color_parts(self, foreground, background):
        foreground_parts = self.split_color_parts(foreground)
        foreground_parts = self.check_invalid_color_parts(foreground_parts)
        background_parts = self.split_color_parts(background)
        background_parts = self.check_invalid_color_parts(background_parts)
        return foreground_parts, background_parts

    def split_color_parts(self, color_parts):
        parts = color_parts.split() if color_parts else []
        return parts

    def is_modifier(self, elem):
        return elem in VALID_COLOR_MODIFIERS

    def sort_color_parts(self, color_parts):
        def comparator(first, second):
            if self.is_modifier(first) and not self.is_modifier(second):
                return 1
            elif not self.is_modifier(first) and self.is_modifier(second):
                return -1
            else:
                return 0

        return sorted(color_parts, key=cmp_to_key(comparator))


class TaskColorizer:
    class Decorator:
        def color_enabled(func):
            @wraps(func)
            def verify_color_enabled(self, *args, **kwargs):
                return func(self, *args, **kwargs) if self.color_enabled else None

            return verify_color_enabled

    def __init__(self, color_config):
        self.color_config = color_config
        self.color_enabled = self.color_config.color_enabled
        self.theme_alt_backgrounds = self.color_config.theme_alt_backgrounds
        self.background_modifier = ""
        self.init_keywords()

    def init_keywords(self):
        try:
            self.keywords = self.color_config.task_config.subtree("color.")["keyword"]
            self.any_keywords_regex = re.compile(
                "(%s)" % "|".join(self.keywords.keys())
            )
        except KeyError:
            self.keywords = []
            self.any_keywords_regex = None

    def has_keywords(self, text):
        return self.any_keywords_regex and self.any_keywords_regex.search(text)

    def extract_keyword_parts(self, text):
        if self.has_keywords(text):
            parts = self.any_keywords_regex.split(text)
            first_part = parts.pop(0)
            return first_part, parts
        return None, None

    def set_background_modifier(self, modifier=""):
        self.background_modifier = (
            modifier if modifier in self.theme_alt_backgrounds else ""
        )

    def add_background_modifier(self, display_attr):
        return display_attr + self.background_modifier

    def make_display_attr(self, display_attr):
        return self.add_background_modifier(display_attr)

    def get_display_attr(self, display_attr):
        return (
            self.make_display_attr(display_attr)
            if self.color_config.has_display_attr(display_attr)
            else None
        )

    @Decorator.color_enabled
    def project_none(self):
        return self.get_display_attr("color.project.none")

    @Decorator.color_enabled
    def project(self, project):
        return self.get_display_attr("color.project.%s" % project)

    @Decorator.color_enabled
    def tag_none(self):
        return self.get_display_attr("color.tag.none")

    @Decorator.color_enabled
    def tag(self, tag):
        custom_value = "color.tag.%s" % tag
        if self.color_config.has_display_attr(custom_value):
            return self.make_display_attr(custom_value)
        elif self.color_config.has_display_attr("color.tagged"):
            return self.make_display_attr("color.tagged")
        return None

    @Decorator.color_enabled
    def uda_none(self, name):
        return self.get_display_attr("color.uda.%s.none" % name)

    @Decorator.color_enabled
    def uda_common(self, name, value):
        custom_value = "color.uda.%s" % name
        if self.color_config.has_display_attr(custom_value):
            return self.make_display_attr(custom_value)
        elif self.color_config.has_display_attr("color.uda"):
            return self.make_display_attr("color.uda")
        return None

    @Decorator.color_enabled
    def uda_string(self, name, value):
        if not value:
            return self.uda_none(name)
        else:
            custom_value = "color.uda.%s.%s" % (name, value)
            if self.color_config.has_display_attr(custom_value):
                return self.make_display_attr(custom_value)
            return self.uda_common(name, value)

    @Decorator.color_enabled
    def uda_numeric(self, name, value):
        return self.uda_string(name, value)

    @Decorator.color_enabled
    def uda_duration(self, name, value):
        return self.uda_string(name, value)

    @Decorator.color_enabled
    def uda_date(self, name, value):
        # TODO: Maybe some special string indicators here?
        return self.uda_common(name, value) if value else self.uda_none(name)

    @Decorator.color_enabled
    def uda_indicator(self, name, value):
        return self.uda_string(name, value)

    @Decorator.color_enabled
    def keyword(self, text):
        return self.get_display_attr("color.keyword.%s" % text)

    @Decorator.color_enabled
    def blocking(self):
        return self.get_display_attr("color.blocking")

    @Decorator.color_enabled
    def due(self, state):
        return self.get_display_attr("color.%s" % state) if state else None

    @Decorator.color_enabled
    def status(self, status):
        if status == "completed" or status == "deleted":
            value = "color.%s" % status
            if self.color_config.has_display_attr(value):
                return self.make_display_attr(value)
        return None

    @Decorator.color_enabled
    def blocked(self, depends):
        return self.get_display_attr("color.blocked")

    @Decorator.color_enabled
    def active(self, active):
        return self.get_display_attr("color.active") if active else None

    @Decorator.color_enabled
    def recurring(self, recur):
        return self.get_display_attr("color.recurring")

    @Decorator.color_enabled
    def scheduled(self, scheduled):
        return self.get_display_attr("color.scheduled") if scheduled else None

    @Decorator.color_enabled
    def until(self, until):
        return self.get_display_attr("color.until") if until else None



import os
try:
    import importlib.util
except:
    import imp

from vit import env, xdg

DEFAULT_VIT_DIR = '~/.vit'

class Loader:
    def __init__(self):
        self.user_config_dir = os.path.expanduser('VIT_DIR' in env.user and env.user['VIT_DIR'] or DEFAULT_VIT_DIR)

        if not os.path.exists(self.user_config_dir):
            xdg_dir = xdg.get_xdg_config_dir(self.user_config_dir, "vit")
            if xdg_dir:
                self.user_config_dir = xdg_dir

    def load_user_class(self, module_type, module_name, class_name):
        module = '%s.%s' % (module_type, module_name)
        filepath = '%s/%s/%s.py' % (self.user_config_dir, module_type, module_name)
        try:
            mod = self.import_from_path(module, filepath)
        except SyntaxError as e:
            raise SyntaxError("User class: %s (%s) -- %s" % (class_name, filepath, e))
        except:
            return None
        return getattr(mod, class_name)

    def import_from_path(self, module, filepath):
        spec = importlib.util.spec_from_file_location(module, filepath)
        mod = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(mod)
        return mod



import urwid

class MultiWidget(urwid.Widget):
    """A widget container that presents one child widget at a time."""

    #: A dict containing all widgets.
    widgets = {}

    #: Name of the current widget.
    current = None

    def __init__(self):
        self.widgets = {}
        self.current = None

    def add_widget(self, name, widget):
        """Adds a widget by name."""
        self.widgets[name] = widget

    def show_widget(self, name):
        assert 0 < len(self.widgets)
        self.current = name
        self._invalidate()

    @property
    def widget_count(self):
        """The function name is pretty much self-explanatory."""
        return len(self.widgets)

    @property
    def current_widget(self):
        """Returns a widget that is currently being rendered. If the widget
        list is empty, it returns None."""
        if self.current and self.current in self.widgets:
            return self.widgets[self.current]
        else:
            return None

    def selectable(self):
        """It appears ``selectable()`` must return ``True`` in order to get any
        key input."""

        return True

    def rows(self, size, focus=False):
        return self.current_widget.rows(size, focus) if self.current_widget is not None else 0

    def render(self, size, focus=False):
        assert self.current_widget is not None

        return self.current_widget.render(size, focus)

    def keypress(self, size, key):
        """Passes key inputs to the current widget. If the current widget is
        ``None`` then it returns the given key input so that
        ``unhandled_input`` function can handle it."""

        if self.current_widget is not None:
            return self.current_widget.keypress(size, key)
        else:
            return key

    def mouse_event(self, size, event, button, col, row, focus):
        if self.current_widget is not None:
            return self.current_widget.mouse_event(
                size, event, button, col, row, focus)
        else:
            return False



from vit.formatter import Formatter
from vit.util import unicode_len

class UdaIndicator(Formatter):
    def format(self, value, task):
        if not value:
            return self.markup_none(self.colorize())
        else:
            indicator = self.formatter.indicator_uda[self.column]
            return (unicode_len(indicator), (self.colorize(value), indicator))

    def colorize(self, value=None):
        return self.colorizer.uda_indicator(self.column, value)



# TODO: Use urwid signals instead of custom event emitter?
class Emitter:
    """Simple event listener/emitter.
    """
    def __init__(self):
        self.listeners = {}

    def listen(self, event, listener):
        if event not in self.listeners:
            self.listeners[event] = []
        self.listeners[event].append(listener)

    def emit(self, event, data=None):
        if event in self.listeners:
            for listener in self.listeners[event]:
                listener(data)



from vit import util
from vit.formatter import List

class Depends(List):
    def format_list(self, depends, task):
        return ','.join(list(map(lambda t: str(util.task_id_or_uuid_short(t)), self.filter_by_blocking_task_uuids(depends)))) if depends else ''

    def colorize(self, depends):
        return self.colorizer.blocked(depends)



from vit.formatter import String
from vit.util import unicode_len

class UdaDuration(String):
    def format(self, duration, task):
        if not duration:
            return self.markup_none(self.colorize())
        return (unicode_len(duration), self.markup_element(duration))
    def colorize(self, duration=None):
        return self.colorizer.uda_duration(self.column, duration)



import datetime
from vit.formatter import DateTime
from vit.util import unicode_len

# TODO: Remove this once tasklib bug is fixed.
from tasklib.serializing import SerializingObject
serializer = SerializingObject({})

class UdaDate(DateTime):
    def format(self, dt, task):
        if not dt:
            return self.markup_none(self.colorize())
        # TODO: Remove this once tasklib bug is fixed.
        # https://github.com/robgolding/tasklib/issues/30
        dt = dt if isinstance(dt, datetime.datetime) else serializer.timestamp_deserializer(dt)
        formatted_date = dt.strftime(self.custom_formatter or self.formatter.report)
        return (unicode_len(formatted_date), (self.colorize(dt), formatted_date))
    def colorize(self, dt=None):
        return self.colorizer.uda_date(self.column, dt)



from vit.formatter.depends import Depends

class DependsCount(Depends):
    def format_list(self, depends, task):
        return '[%d]' % len(self.filter_by_blocking_task_uuids(depends)) if depends else ''



from functools import reduce

class KeyCacheError(Exception):
    pass

class KeyCache:
    def __init__(self, keybindings):
        self.keybindings = keybindings
        self.cached_keys = ''
        self.build_multi_key_cache()

    def get(self, key=None):
        return '%s%s' % (self.cached_keys, key) if key else self.cached_keys

    def set(self, keys=''):
        self.cached_keys = keys

    def is_keybinding(self, keys):
        return keys in self.keybindings

    def sort_keybindings_by_len(self, keybindings, min_len=1):
        max_key_length = len(max(keybindings, key=len))
        def reducer(accum, key_length):
            accum.append([v for k, v in enumerate(keybindings) if len(keybindings[k]) == key_length])
            return accum
        sorted_keybindings = reduce(reducer, range(max_key_length, min_len, -1), [])
        return list(filter(None, sorted_keybindings))

    def get_non_modified_keybindings(self):
        return [k for k in self.keybindings if self.keybindings[k]['has_special_keys'] or not self.keybindings[k]['has_modifier']]

    def add_keybinding_to_key_cache(self, to_cache, keybinding, existing_keybindings, key_cache):
        if to_cache in existing_keybindings:
            raise KeyCacheError("Invalid key binding '%s', '%s' already used in another key binding" % (keybinding, to_cache))
        else:
            key_cache[to_cache] = True

    def build_multi_key_cache(self):
        keybindings = self.get_non_modified_keybindings()
        sorted_keybindings = self.sort_keybindings_by_len(keybindings)
        def sorted_keybindings_reducer(key_cache, keybinding_list):
            def keybinding_list_reducer(_, keybinding):
                keys = list(keybinding)
                keys.pop()
                def keybinding_reducer(processed_keys, key):
                    processed_keys.append(key)
                    to_cache = ''.join(processed_keys)
                    self.add_keybinding_to_key_cache(to_cache, keybinding, keybindings, key_cache)
                    return processed_keys
                reduce(keybinding_reducer, keys, [])
            reduce(keybinding_list_reducer, keybinding_list, [])
            return key_cache
        self.multi_key_cache = reduce(sorted_keybindings_reducer, sorted_keybindings, {})



from functools import reduce

try:
    import configparser
except ImportError:
    import ConfigParser as configparser

import os
import re

from vit import util

BASE_DIR = os.path.dirname(os.path.realpath(__file__))
BRACKETS_REGEX = re.compile("[<>]")
DEFAULT_KEYBINDINGS_SECTIONS = ('global', 'navigation', 'command', 'report')
CONFIG_SPECIAL_KEY_SUBSTITUTIONS = {
    'colon': ':',
    'equals': '=',
    'space': ' ',
    'semicolon': ';',
}

class KeybindingError(Exception):
    pass

class KeybindingParser:
    def __init__(self, loader, config, action_registry):
        self.loader = loader
        self.config = config
        self.action_registry = action_registry
        self.actions = self.action_registry.get_actions()
        self.noop_action_name = self.action_registry.make_action_name(self.action_registry.noop_action_name)
        self.default_keybinding_name = self.config.get('vit', 'default_keybindings')
        self.default_keybindings = configparser.ConfigParser()
        self.default_keybindings.optionxform=str
        self.sections = DEFAULT_KEYBINDINGS_SECTIONS
        self.keybindings = {}
        self.multi_key_cache = {}

    def is_keybinding(self, keys):
        return keys in self.keybindings

    def items(self, section):
        try:
            return self.default_keybindings.items(section)
        except configparser.NoSectionError:
            return []

    def load_default_keybindings(self):
        name = self.default_keybinding_name
        template = '%s/keybinding/%s.ini'
        user_keybinding_file = template % (self.loader.user_config_dir, name)
        keybinding_file = template % (BASE_DIR, name)
        if util.file_readable(user_keybinding_file):
            self.default_keybindings.read(user_keybinding_file)
        elif util.file_readable(keybinding_file):
            self.default_keybindings.read(keybinding_file)
        else:
            raise KeybindingError("default_keybindings setting '%s' invalid, file not found" % name)
        for section in self.sections:
            bindings = self.items(section)
            self.add_keybindings(bindings)

    def keybinding_special_keys_substitutions(self, value):
        is_special_key = value in CONFIG_SPECIAL_KEY_SUBSTITUTIONS
        if is_special_key:
            value = CONFIG_SPECIAL_KEY_SUBSTITUTIONS[value]
        return value, is_special_key

    def parse_keybinding_keys(self, keys):
        has_modifier = bool(re.match(BRACKETS_REGEX, keys))
        def reducer(accum, char):
            if char == '<':
                accum['in_brackets'] = True
                accum['bracket_string'] = ''
            elif char == '>':
                accum['in_brackets'] = False
                value, is_special_key = self.keybinding_special_keys_substitutions(accum['bracket_string'].lower())
                accum['keys'] += value
                accum['has_special_keys'] = is_special_key
            else:
                if accum['in_brackets']:
                    accum['bracket_string'] += char
                else:
                    accum['keys'] += char
            return accum
        accum = reduce(reducer, keys, {
            'keys': '',
            'in_brackets': False,
            'bracket_string': '',
            'has_special_keys': False,
        })
        return accum['keys'], has_modifier, accum['has_special_keys']

    def parse_keybinding_value(self, value, replacements={}):
        def reducer(accum, char):
            if char == '<':
                accum['in_brackets'] = True
                accum['bracket_string'] = ''
            elif char == '>':
                accum['in_brackets'] = False
                value, is_special_key = self.keybinding_special_keys_substitutions(accum['bracket_string'].lower())
                accum['keybinding'].append(value)
            elif char == '{':
                accum['in_variable'] = True
                accum['variable_string'] = ''
            elif char == '}':
                accum['in_variable'] = False
                if accum['variable_string'] in self.actions:
                    accum['action_name'] = accum['variable_string']
                else:
                    for replacement in replacements:
                        args = replacement['match_callback'](accum['variable_string'])
                        if isinstance(args, list):
                            accum['keybinding'].append((replacement['replacement_callback'], args))
                            return accum
                    raise ValueError("unknown config variable '%s'" % accum['variable_string'])
            else:
                if accum['in_brackets']:
                    accum['bracket_string'] += char
                elif accum['in_variable']:
                    accum['variable_string'] += char
                else:
                    accum['keybinding'].append(char)
            return accum
        accum = reduce(reducer, value, {
            'keybinding': [],
            'action_name': None,
            'in_brackets': False,
            'bracket_string': '',
            'in_variable': False,
            'variable_string': '',
        })
        return accum['keybinding'], accum['action_name']

    def validate_parsed_value(self, key_groups, bound_keys, action_name):
        if bound_keys and action_name:
            raise KeybindingError("keybindings '%s' unsupported configuration: ACTION_ variables must be used alone." % key_groups)

    def is_noop_action(self, keybinding):
        return True if 'action_name' in keybinding and keybinding['action_name'] == self.noop_action_name else False

    def filter_noop_actions(self, keybindings):
        return {keys:value for (keys, value) in keybindings.items() if not self.is_noop_action(keybindings[keys])}

    def get_keybindings(self):
        return self.keybindings

    def add_keybindings(self, bindings=[], replacements={}):
        for key_groups, value in bindings:
            bound_keys, action_name = self.parse_keybinding_value(value, replacements)
            self.validate_parsed_value(key_groups, bound_keys, action_name)
            for keys in key_groups.strip().split(','):
                parsed_keys, has_modifier, has_special_keys = self.parse_keybinding_keys(keys)
                self.keybindings[parsed_keys] = {
                    'label': keys,
                    'has_modifier': has_modifier,
                    'has_special_keys': has_special_keys,
                }
                if action_name:
                    self.keybindings[parsed_keys]['action_name'] = action_name
                else:
                    self.keybindings[parsed_keys]['keys'] = bound_keys
        self.keybindings = self.filter_noop_actions(self.keybindings)
        return self.get_keybindings()




theme = [
    ('list-header', '', '', '', '', ''),
    ('list-header-column', 'black', 'light gray', '', 'black', 'light gray'),
    ('list-header-column-separator', 'black', 'light gray', '', 'black', 'light gray'),
    ('striped-table-row', 'white', 'dark gray', '', 'white', 'g27'),
    ('reveal focus', 'black', 'dark cyan', 'standout', 'black', 'dark cyan'),
    ('message status', 'white', 'dark blue', 'standout', 'white', 'dark blue'),
    ('message error', 'white', 'dark red', 'standout', 'white', 'dark red'),
    ('status', 'dark magenta', 'black', '', 'dark magenta', 'black'),
    ('flash off', 'black', 'black', 'standout', 'black', 'black'),
    ('flash on', 'white', 'black', 'standout', 'white', 'black'),
    ('pop_up', 'white', 'black', '', 'white', 'black'),
    ('button action', 'white', 'light red', '', 'white', 'light red'),
    ('button cancel', 'black', 'light gray', '', 'black', 'light gray'),
]



from vit.formatter import DateTime

class Due(DateTime):
    def get_due_state(self, due, task):
        return self.formatter.get_due_state(due, task)

    def colorize(self, due, task):
        return self.colorizer.due(self.get_due_state(due, task))



import re
from setuptools import setup
from os import path

DEFAULT_BRANCH = "2.x"
BASE_GITHUB_URL = "https://github.com/vit-project/vit/blob"

MARKUP_LINK_REGEX = r"\[([^]]+)\]\(([\w]+\.md)\)"
FILE_DIR = path.dirname(path.abspath(path.realpath(__file__)))

with open(path.join(FILE_DIR, 'README.md')) as f:
    readme_contents = f.read()
    markup_link_substitution = '[\\1](%s/%s/\\2)' % (BASE_GITHUB_URL, DEFAULT_BRANCH)
    README = re.sub(MARKUP_LINK_REGEX, markup_link_substitution, readme_contents, flags=re.MULTILINE)

with open(path.join(FILE_DIR, 'requirements.txt')) as f:
    INSTALL_PACKAGES = f.read().splitlines()

with open(path.join(FILE_DIR, 'vit', 'version.py')) as f:
    VERSION = re.match(r"^VIT = '([\w\.]+)'$", f.read().strip())[1]

setup(
    name='vit',
    packages=['vit', 'vit.config', 'vit.formatter', 'vit.keybinding', 'vit.theme'],
    description="Visual Interactive Taskwarrior full-screen terminal interface",
    long_description=README,
    long_description_content_type='text/markdown',
    install_requires=INSTALL_PACKAGES,
    version=VERSION,
    url='https://github.com/vit-project/vit',
    author='Chad Phillips',
    author_email='chad@apartmentlines.com',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: MIT License',
        'Environment :: Console',
        'Programming Language :: Python',
        'Natural Language :: English',
        'Topic :: Text Processing :: General',
    ],
    keywords=[
        'taskwarrior',
        'console',
        'tui',
        'text-user-interface',
    ],
    tests_require=[
    ],
    entry_points = {
        'console_scripts': [
            'vit=vit.command_line:main',
        ],
    },
    include_package_data=True,
    python_requires='>=3.7',
    zip_safe=False
)



import unittest

from pprint import pprint
from vit.list_batcher import ListBatcher

DEFAULT_BATCH_FROM = list('abcdefghijklmnopqrstuvwxyz')

def default_batcher():
    batch_to = []
    return ListBatcher(DEFAULT_BATCH_FROM, batch_to), batch_to

class TestListBatcher(unittest.TestCase):

    def test_batch_default_batch(self):
        batcher, batch_to = default_batcher()
        complete = batcher.add()
        self.assertEqual(complete, True)
        self.assertEqual(batcher.get_last_position(), len(DEFAULT_BATCH_FROM))
        self.assertEqual(batch_to, DEFAULT_BATCH_FROM)

    def test_batch_1(self):
        batcher, batch_to = default_batcher()
        complete = batcher.add(1)
        self.assertEqual(complete, False)
        self.assertEqual(batcher.get_last_position(), 1)
        self.assertEqual(batch_to, ['a'])

    def test_batch_5(self):
        batcher, batch_to = default_batcher()
        complete = batcher.add(5)
        self.assertEqual(complete, False)
        self.assertEqual(batcher.get_last_position(), 5)
        self.assertEqual(batch_to, ['a', 'b', 'c', 'd', 'e'])

    def test_batch_1_and_1(self):
        batcher, batch_to = default_batcher()
        batcher.add(1)
        complete = batcher.add(1)
        self.assertEqual(complete, False)
        self.assertEqual(batcher.get_last_position(), 2)
        self.assertEqual(batch_to, ['a', 'b'])

    def test_batch_1_and_rest(self):
        batcher, batch_to = default_batcher()
        batcher.add(1)
        complete = batcher.add(0)
        self.assertEqual(complete, True)
        self.assertEqual(batcher.get_last_position(), len(DEFAULT_BATCH_FROM))
        self.assertEqual(batch_to, DEFAULT_BATCH_FROM)

    def test_batch_batch_size_greater_than_default(self):
        batcher, batch_to = default_batcher()
        complete = batcher.add(100000)
        self.assertEqual(complete, True)
        self.assertEqual(batcher.get_last_position(), len(DEFAULT_BATCH_FROM))
        self.assertEqual(batch_to, DEFAULT_BATCH_FROM)

    def test_batch_add_on_completed(self):
        batcher, batch_to = default_batcher()
        complete = batcher.add()
        self.assertEqual(complete, True)
        complete = batcher.add()
        self.assertEqual(complete, True)
        self.assertEqual(batcher.get_last_position(), len(DEFAULT_BATCH_FROM))
        self.assertEqual(batch_to, DEFAULT_BATCH_FROM)

    def test_batch_5_with_formatter(self):
        def formatter(partial, start_idx):
            return ['before'] + [row * 2 for row in partial] + ['after']
        batch_to = []
        batcher = ListBatcher(DEFAULT_BATCH_FROM, batch_to, batch_to_formatter=formatter)
        complete = batcher.add(5)
        self.assertEqual(complete, False)
        self.assertEqual(batcher.get_last_position(), 5)
        self.assertEqual(batch_to, ['before', 'aa', 'bb', 'cc', 'dd', 'ee', 'after'])

if __name__ == '__main__':
    unittest.main()
